from base64 import b64encode
from OpenSSL import crypto

STP_EMPRESA = None
STP_PRIVKEY = None
STP_PRIVKEY_PASSPHRASE = None
STP_PREFIJO = None
SIGN_DIGEST = 'RSA-SHA256'
ACTUALIZA_CLIENT = None
DEBUG_MODE = False
HISTORY = None


def _join_fields(obj, fieldnames):
    fields = []
    for name in fieldnames:
        if name in ['monto', 'iva'] and not getattr(obj, name) in ['', None]:
            field = float(getattr(obj, name))
            field = f'{field:.2f}'
        else:
            field = getattr(obj, name) or ''
        fields.append(str(field))
    return ('||' + '|'.join(fields) + '||').encode('utf-8')


def _validate(field, field_value, validation, validation_value):
    """
        Evalua si el valor de un campo es válido y devuelve un mensaje de error
    :param field: Nombre del campo
    :param field_value: Valor del campo
    :param validation: Tipo de validación a realizar
    :param validation_value: Valor a validar
    :return:
    """
    if validation == 'required' and validation_value and field_value is None:
        return 'Field {} is required'.format(field)
    if validation == 'maxLength' and validation_value < len(str(field_value)):
        return 'Length of field {} must be lower than {}'\
            .format(field, validation_value)
    return None


class Resource:
    __fieldnames__ = None
    __object__ = None
    __type__ = None
    __validations__ = None
    _defaults = {}

    def __init__(self, **kwargs):
        for default, value in self._defaults.items():
            if default not in kwargs:
                if callable(value):
                    kwargs[default] = value()
                else:
                    kwargs[default] = value
        self.__object__ = ACTUALIZA_CLIENT.get_type(self.__type__)(**kwargs)
        self.empresa = STP_EMPRESA
        self.firma = None

    def __dir__(self):
        return dir(super(Resource, self)) + dir(self.__object__)

    def __dict__(self):
        return {r: self.__getattr__(r) for r in self.__fieldnames__}

    def __eq__(self, other):
        return all(getattr(self, name) ==
                   getattr(other, name) for name in self.__fieldnames__)

    def __getattr__(self, item):
        if item.startswith('_'):
            return self.__getattribute__(item)
        return getattr(self.__object__, item)

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        indent = ' ' * 4
        rv = f'{self.__class__.__name__}(\n'
        for name in self.__fieldnames__:
            rv += f'{indent}{name}={repr(getattr(self, name))},\n'
        rv += ')'
        return rv

    def __setattr__(self, key, value):
        if key.startswith('_'):
            super(Resource, self).__setattr__(key, value)
        else:
            setattr(self.__object__, key, value)

    def __str__(self):
        return self.__object__.__str__()

    @property
    def _joined_fields(self):
        return _join_fields(self, self.__fieldnames__)

    def _compute_signature(self):
        signature = crypto.sign(STP_PRIVKEY, self._joined_fields, SIGN_DIGEST)
        return b64encode(signature).decode('ascii')

    def _is_valid_field(self, field):
        """
            Obtiene el campo a ser evaluado y el diccionario de validaciones
            que deban hacerse
        :param field: Campo a evaluar
        :return: Lista de errores encontrados en este campo
        """
        # Validaciones que aplican en este campo
        vals = self.__validations__[field]
        return [_validate(field, getattr(self, field), r, vals[r])
                for r in vals]

    def _is_valid(self):
        """
            Por todos los campos a ser validados, ejecuta la función
            _is_valid_field y devuelve todos los errores
        :return: None si no hay errores, de otra forma lanza una
            excepción con la lista de errores
        """
        errors = list(filter((lambda x: x is not None),
                             [error for errors in
                              map((lambda r: self._is_valid_field(r)),
                                  self.__validations__)
                              for error in errors]))
        if len(errors) > 0:
            raise ValueError(",".join(errors))
        return None

    def _invoke_method(self, method):
        res = ACTUALIZA_CLIENT.service[method](self.__object__)
        return res



from clabe.banks import BankCode

from stpmex.types import Institucion


def spei_to_stp_bank_code(spei_code: str) -> Institucion:
    try:
        bank_name = BankCode(spei_code)
        return Institucion[bank_name.name]
    except (ValueError, KeyError):
        return None


def stp_to_spei_bank_code(stp_code: int) -> str:
    try:
        bank_name = Institucion(stp_code)
        return BankCode[bank_name.name].value
    except (ValueError, KeyError):
        return None



"""
Configuraciones iniciales para utilizar el cliente posteriormente
"""
from zeep import Client
from zeep.transports import Transport
from OpenSSL import crypto
from requests import Session
from .ordenes import Orden
from .base import STP_EMPRESA, STP_PREFIJO, STP_PRIVKEY, STP_PRIVKEY_PASSPHRASE

DEFAULT_WSDL = ('https://demo.stpmex.com:7024/speidemo/webservices/SpeiActual'
                'izaServices?wsdl')


def configure(empresa: str, priv_key: str, priv_key_passphrase: str,
              prefijo: int, wsdl_path: str = DEFAULT_WSDL, proxy: str = None,
              proxy_user: str = None, proxy_password: str = None):
    """
    Configura las credenciales y parámetros necesarios para poder hacer
    peticiones a STP
    :param empresa: Nombre de la empresa
    :param priv_key: Contenido de la llave privada
    :param priv_key_passphrase: Password de la llave privada
    :param prefijo: Prefijo en STP
    :param wsdl_path: URL del WSDL a utilizar
    :param proxy: Si es necesario, se puede especificar un proxy
    :param proxy_user: Usuario del proxy
    :param proxy_password: Contraseña del proxy
    :return:
    """
    base.STP_EMPRESA = empresa
    base.STP_PRIVKEY = crypto.load_privatekey(crypto.FILETYPE_PEM, priv_key,
                                              priv_key_passphrase.
                                              encode('ascii'))
    base.STP_PRIVKEY_PASSPHRASE = priv_key_passphrase
    base.STP_PREFIJO = prefijo
    base.WSDL_PATH = wsdl_path

    if proxy is not None:
        session = Session()
        session.proxies = {
            'https': f"https://{proxy_user}:{proxy_password}@{proxy}",
            'http':  f"http://{proxy_user}:{proxy_password}@{proxy}"
        }
        base.ACTUALIZA_CLIENT = Client(wsdl_path,
                                       transport=Transport(session=session))
    else:
        base.ACTUALIZA_CLIENT = Client(wsdl_path)



from enum import Enum


class AccountType(Enum):
    DEBIT_CARD = 3
    PHONE_NUMBER = 10
    CLABE = 40


class Prioridad(Enum):
    normal = 0
    alta = 1


class Institucion(Enum):
    BANCOMEXT = 37006
    BANOBRAS = 37009
    BANJERCITO = 37019
    NAFIN = 37135
    BANSEFI = 37166
    HIPOTECARIAFEDERAL = 37168
    BANAMEX = 40002
    BBVA_BANCOMER = 40012
    SANTANDER = 40014
    HSBC = 40021
    BAJIO = 40030
    INBURSA = 40036
    INTERACCIONES = 40037
    MIFEL = 40042
    SCOTIABANK = 40044
    BANREGIO = 40058
    INVEX = 40059
    BANSI = 40060
    AFIRME = 40062
    BANORTE = 40072
    INVESTA_BANK = 40102
    AMERICANEXPRESS = 40103
    BAMSA = 40106
    TOKYO = 40108
    JPMORGAN = 40110
    BMONEX = 40112
    VEPORMAS = 40113
    DEUTSCHE = 40124
    CREDITSUISSE = 40126
    AZTECA = 40127
    AUTOFIN = 40128
    BARCLAYS = 40129
    COMPARTAMOS = 40130
    FAMSA = 40131
    MULTIVA_BANCO = 40132
    ACTINVER = 40133
    INTERCAM = 40136
    BANCOPPEL = 40137
    ABCCAPITAL = 40138
    CONSUBANCO = 40140
    VOLKSWAGEN = 40140
    CIBANCO = 40143
    BBASE = 40145
    BANKAOOL = 40147
    PAGATODO = 40148
    INMOBILIARIO = 40150
    DONDE = 40151
    BACREA = 40152
    BANCO_FINTERRA = 40154
    ICBC = 40155
    SABADELL = 40156
    MIZUHO_BANK = 40156
    MONEXCB = 90600
    GBP = 90601
    MASARI = 90602
    VALUE = 90605
    ESTRUCTURADORES = 90606
    VECTOR = 90608
    MULTIVA_CBOLSA = 90613
    ACCIVAL = 90614
    FINAMEX = 90616
    VALMEX = 90617
    PROFUTURO = 90620
    CB_ACTINVER = 90621
    SKANDIAVIDA = 90623
    CBDEUTSCHE = 90626
    ZURICH = 90627
    ZURICHVI = 90628
    CBINTERCAM = 90630
    CI_BOLSA = 90631
    FINCOMUN = 90634
    HDI_SEGUROS = 90636
    ORDER = 90637
    AKALA = 90638
    JPMORGANCB = 90640
    REFORMA = 90642
    STP = 90646
    EVERCORE = 90648
    OSKNDIA = 90649
    ASEA = 90652
    KUSPIT = 90653
    SOFIEXPRESS = 90655
    UNAGRA = 90656
    ASP_INTEGRA_OPC = 90659
    LIBERTAD = 90670
    HUASTECAS = 90671
    AXA = 90674
    CAJA_POP_MEXICA = 90677
    SURA = 90678
    FND = 90679
    CRISTOBAL_COLON = 90680
    PRINCIPAL = 90681
    CAJA_TELEFONIST = 90683
    TRANSFER = 90684
    FONDO_FIRA = 90685
    INVERCAP = 90686
    INFONAVIT = 90687
    FOMPED = 90689
    CLS = 90901
    INDEVAL = 90902



"""
CLI para enviar órdenes a STP.
"""
import argparse
import json
import stpmex
from stpmex import Orden
from stpmex.types import Institucion

DEFAULT_FILE_NAME = 'stp.config'


def config():
    """
    Crea un archivo de configuración con las credenciales de STP para poder
    enviar órdenes posteriormente. PRECAUCIÓN: El archivo se guarda sin
    cifrar y en la ubicación donde se encuentre la línea de comandos en ese
    momento.
    :return:
    """
    print('Configuring connection')
    print('It will create a stp.config file')

    pk_file = input('Route to private key: ')
    with open(pk_file) as fp:
        pk_value = fp.read()
    configuration = {
        'wsdl': input('WSDL: '),
        'private_key': pk_value,
        'pkey_passphrase': input('Private key passphrase: '),
        'empresa': input('Empresa: '),
        'prefijo': input('Prefijo: ')
    }
    use_proxy = input('Wish to use a proxy? (y/n): ').lower() == 'y'
    if use_proxy:
        configuration['proxy'] = input('Proxy: ')
        configuration['user'] = input('User: ')
        configuration['password'] = input('Password: ')

    with open(DEFAULT_FILE_NAME, 'w') as outfile:
        json.dump(configuration, outfile)

    print("Done...")


def order():
    """
    Crea una nueva orden para enviar a STP
    :return:
    """
    try:
        with open(DEFAULT_FILE_NAME, 'r') as f:
            config = json.load(f)
    except IOError:
        print("No configuration file found, use first: stpmex config")
        return

    stpmex.configure(wsdl_path=config['wsdl'], empresa=config['empresa'],
                     priv_key=config['private_key'],
                     priv_key_passphrase=config['pkey_passphrase'],
                     prefijo=config['prefijo'],
                     proxy=None if 'proxy' not in config else config['proxy'],
                     proxy_user=None if 'user' not in config
                     else config['user'],
                     proxy_password=None if 'password' not in config
                     else config['password'],
                     )

    print("Connection established")

    order = Orden()
    order.institucionOperante = Institucion.STP.value
    order.nombreBeneficiario = input('Nombre del beneficiario: ')
    order.cuentaBeneficiario = input('CLABE del beneficiario: ')
    order.institucionContraparte = input('Institución de contraparte: ')
    order.conceptoPago = input('Concepto de pago: ')
    order.monto = float(input('Monto: '))  # I assume you won't put a str here
    rfc = input('RFC o CURP de beneficiario (opcional): ')
    if rfc:
        order.rfcCurpBeneficiario = rfc
    num_reference = input('Referencia númerica (opcional): ')
    if num_reference:
        order.referenciaNumerica = num_reference
    track_code = input('Clave de rastreo (opcional): ')
    if track_code:
        order.claveRastreo = track_code

    print("Sending order....")
    r = order.registra()

    print(r)
    print("Finished")


def main():
    """
    Función principal, recibe como argumento la función a utilizar
    :return:
    """
    parser = argparse.ArgumentParser(description='Creates an order to STP')
    parser.add_argument('option', choices=['config', 'order'],
                        help="'config' to configure the client,"
                             "'order' for creating a new order")
    args = parser.parse_args()
    if args.option == 'config':
        config()
    if args.option == 'order':
        order()


if __name__ == '__main__':
    main()



import random
import time

from .base import Resource
from .types import AccountType, Prioridad

ORDEN_FIELDNAMES = """
    institucionContraparte
    empresa
    fechaOperacion
    folioOrigen
    claveRastreo
    institucionOperante
    monto
    tipoPago
    tipoCuentaOrdenante
    nombreOrdenante
    cuentaOrdenante
    rfcCurpOrdenante
    tipoCuentaBeneficiario
    nombreBeneficiario
    cuentaBeneficiario
    rfcCurpBeneficiario
    emailBeneficiario
    tipoCuentaBeneficiario2
    nombreBeneficiario2
    cuentaBeneficiario2
    rfcCurpBeneficiario2
    conceptoPago
    conceptoPago2
    claveCatUsuario1
    claveCatUsuario2
    clavePago
    referenciaCobranza
    referenciaNumerica
    tipoOperacion
    topologia
    usuario
    medioEntrega
    prioridad
    iva
    """.split()

ORDEN_DEFAULTS = dict(
    rfcCurpBeneficiario='ND',
    tipoPago=1,
    tipoCuentaBeneficiario=AccountType.CLABE.value,
    topologia='T',
    medioEntrega=3,
    prioridad=Prioridad.alta.value,
    claveRastreo=lambda: f'CR{int(time.time())}',
    referenciaNumerica=lambda: random.randint(10 ** 6, 10 ** 7)
)

VALIDATIONS = dict(
    nombreBeneficiario=dict(
        required=True,
        maxLength=39
    ),
    claveRastreo=dict(
        required=True
    ),
    conceptoPago=dict(
        required=True
    ),
    referenciaNumerica=dict(
        required=True,
        maxLength=7
    )
)


class Orden(Resource):
    __fieldnames__ = ORDEN_FIELDNAMES
    __type__ = 'ns0:ordenPagoWS'
    __validations__ = VALIDATIONS
    _id = None
    _defaults = ORDEN_DEFAULTS

    def registra(self):
        self._is_valid()
        self.firma = self._compute_signature()
        resp = self._invoke_method('registraOrden')
        self._id = resp.id
        return resp



import setuptools

requirements = [
    'zeep',
    'pyopenssl',
    'clabe'
]

with open('README.md', 'r') as f:
    long_description = f.read()

setuptools.setup(
    name='stpmex',
    version='0.0.8',
    author='Cuenca',
    author_email='dev@cuenca.io',
    description='Integration to stpmex.com',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/cuenca-mx/stpmex-python',
    packages=setuptools.find_packages(),
    install_requires=requirements,
    extras_require={
        'dev': [
            'pytest>=3',
            'ipython',
            'ipdb',
            'pycodestyle',
            'pytest',
            'vcrpy'
        ]
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    entry_points={
        'console_scripts': [
            'stpmex = stpmex.__main__:main'
        ]
    }
)
